@@ -1,14 +1,14 @@
 import re
-import string
 import datetime
 import simplejson as json
-import pytz
 
 from jinja2 import TemplateNotFound
-from flask import g, request, Response, abort
+from paqforms import variable_decode
+from flask import g, request, Response, abort, render_template
 from flask.ext.routes import route
 
 from .view import View
+from .utils import dashify, Paginator
 
 
 class ModelView(View):
@@ -70,28 +70,28 @@ class ModelView(View):
     @route('/search.json', endpoint='{view}:search_json', methods=('GET', 'POST'))
     @route('/search.html', endpoint='{view}:search_html', methods=('GET', 'POST'))
     def search(self):
-        return self.do_index('search')
+        return self._index('search')
 
 
     @route('/', endpoint='{view}:index', methods=('GET', 'POST'))
     @route('/index.json', endpoint='{view}:index_json', methods=('GET', 'POST'))
     @route('/index.html', endpoint='{view}:index_html', methods=('GET', 'POST'))
     def index(self):
-        return self.do_index('index')
+        return self._index('index')
 
 
     @route('/own/', endpoint='{view}:own', methods=('GET', 'POST'))
     @route('/own/index.json', endpoint='{view}:own_json', methods=('GET', 'POST'))
     @route('/own/index.html', endpoint='{view}:own_html', methods=('GET', 'POST'))
     def own(self):
-        return self.do_index('own')
+        return self._index('own')
 
 
     @route('/favs/', endpoint='{view}:favs', methods=('GET', 'POST'))
     @route('/favs/index.json', endpoint='{view}:favs_json', methods=('GET', 'POST'))
     @route('/favs/index.html', endpoint='{view}:favs_html', methods=('GET', 'POST'))
     def favs(self):
-        return self.do_index('favs')
+        return self._index('favs')
 
 
     @route('/<id:id>/detail', endpoint='{view}:detail')
@@ -101,29 +101,29 @@ class ModelView(View):
     @route('/<id:id>/detail.html', endpoint='{view}:detail_html')
     @route('/<id:id>/card.html', endpoint='{view}:card_html')
     def detail(self, id):
-        return self.do_detail(id)
+        return self._detail(id)
 
 
     @route('/create', methods=('GET', 'POST'))
     def create(self):
-        return self.do_create()
+        return self._create()
 
 
     @route('/<id:id>/edit', methods=('GET', 'POST'))
     def edit(self, id):
-        return self.do_edit(id)
+        return self._edit(id)
 
 
     @route('/delete', methods=('POST',))
     @route('/<id:id>/delete', methods=('POST',))
     def delete(self, id=None):
-        return self.do_delete(id)
+        return self._delete(id)
 
 
     @route('/archive', methods=('POST',))
     @route('/<id:id>/archive', methods=('POST',))
     def archive(self, id):
-        return self.do_archive(id)
+        return self._archive(id)
 
 
     # ACTIONS (ROUTES + IMPLEMENTATIONS) ---------------------------------------
@@ -134,19 +134,17 @@ class ModelView(View):
 
 
     # IMPLEMENTATIONS ----------------------------------------------------------
-    def do_index(self, index):
+    def _index(self, index):
         if not self.allows(index):
             raise abort(403)
 
         filters, filterform = self.do_filter(self.filterform, index)
         sorts, sortform = self.do_sort(self.sortform, index)
-
         cursor = self.do_cursor(filters, sorts)
-        self.after_cursor(cursor)
         paginator, redirect = self.do_paginate(cursor, index)
         if redirect:
             return redirect
-        data = self.after_index(index, paginator)
+        data = self.do_index(paginator, index)
 
         try:
             if request.path.endswith('.json'):
@@ -191,7 +189,7 @@ class ModelView(View):
             raise abort(404)
 
 
-    def do_detail(self, id):
+    def _detail(self, id):
         model, data = self.detail_model(id)
         if not self.allows('detail', model):
             raise abort(403)
@@ -235,17 +233,14 @@ class ModelView(View):
             raise abort(404)
 
 
-    def do_create(self):
+    def _create(self):
         model, data = self.create_model()
         if not self.allows('create', model):
             raise abort(403)
 
         form = self.create_form_class(model, request.form)
         if form.ok:
-            self.before_create(model, form)
-            self.do_save(model)
-            self.rebind_files(model)
-            self.after_create(model, form)
+            self.do_create(model, form)
             return self.autoredirect(model.id)
 
         return self.render('create',
@@ -255,17 +250,14 @@ class ModelView(View):
         )
 
 
-    def do_edit(self, id):
+    def _edit(self, id):
         model, data = self.edit_model(id)
         if not self.allows('edit', model):
             raise abort(403)
 
         form = self.edit_form_class(model, request.form)
         if form.ok:
-            self.before_edit(model, form)
-            model.edited_on = utcnow()
-            self.do_save(model)
-            self.after_edit(model, form)
+            self.do_edit(model, form)
             return self.autoredirect(model.id)
 
         return self.render('edit',
@@ -275,6 +267,183 @@ class ModelView(View):
         )
 
 
+    def _delete(self, id):
+        model, data = self.detail_model(id=id)
+        if not self.allows('delete', model):
+            raise abort(403)
+        self.do_delete([model])
+        flash_html = render_template('alerts.html')
+        redirect = request.args.get('redirect')
+        return jsonify(
+            data_html = '',
+            flash_html = flash_html,
+            redirect = redirect,
+            **data
+        )
+
+
+    def _archive(self, id):
+        model, data = self.detail_model(id)
+        if not g.user.allowed(self.endpoint + '.archive', model):
+            raise abort(403)
+        self.do_archive([model])
+        flash_html = render_template('alerts.html')
+        redirect = request.args.get('redirect')
+        return jsonify(
+            data_html = '',
+            flash_html = flash_html,
+            redirect = redirect,
+            **data
+        )
+
+
+    # HOOKS --------------------------------------------------------------------
+    def do_filter(self, form, index):
+        if form:
+            form_class = form.__class__
+            default_value = form.feed_value.copy()
+
+            action = request.form.get('action') or request.args.get('action')
+            inputfilters = variable_decode(request.args).get('filters', {})
+            inputfilters.update(variable_decode(request.form).get('filters', {}))
+
+            if action == 'filters.apply':
+                form = form_class({}, inputfilters, submit=True)
+                if not form.has_error:
+                    self.session.setdefault(index, {})['filters'] = inputfilters
+            elif action == 'filters.reset':
+                form = form_class(default_value)
+                try: del self.session[index]['filters']
+                except: pass
+            elif inputfilters:
+                form = form_class({}, inputfilters)
+            else:
+                # Use session
+                data = self.session.get(index, {}).get('filters', default_value)
+                form = form_class({}, data)
+
+            filters = {} if form.has_error else self.get_filters(form)
+        else:
+            filters = {}
+            form = None
+
+        self.force_filters(index, filters)
+
+        return filters, form
+
+
+    def do_sort(self, form, index):
+        if form:
+            form_class = form.__class__
+            default_value = form.feed_value.copy()
+
+            action = request.form.get('action') or request.args.get('action')
+            inputsorts = variable_decode(request.args).get('sorts', {})
+            inputsorts.update(variable_decode(request.form).get('sorts', {}))
+            fields = request.form.get('fields') or request.args.get('fields')
+
+            if action == 'sorts.apply':
+                form = form_class({}, inputsorts, submit=True)
+                if request.form.get('fields'):
+                    form.fields = OrderedDict([(name, form.fields[name]) for name in request.form.get('fields').split(' ')])
+                if not form.has_error:
+                    self.session.setdefault(index, {})['sorts'] = inputsorts
+                    self.session[index]['fields'] = fields
+            elif action == 'sorts.reset':
+                form = form_class(default_value)
+                try: del self.session[index]['sorts']
+                except: pass
+                try: del self.session[index]['fields'];
+                except: pass
+            elif inputsorts:
+                form = form_class({}, inputsorts)
+            else:
+                # Use session
+                data = self.session.get(index, {}).get('sorts', default_value)
+                form = form_class({}, data)
+                if self.session.get(index, {}).get('fields'):
+                    form.fields = OrderedDict([(name, form.fields[name]) for name in self.session.get(index, {}).get('fields').split(' ')])
+
+            sorts = [] if form.has_error else self.get_sorts(form)
+        else:
+            sorts = []
+            form = None
+
+        return sorts, form
+
+
+    def force_filters(self, index, filters):
+        pass
+
+
+    def do_cursor(self, filters, sorts):
+        raise Exception('Define `{}.do_cursor` method'.format(self.__class__.__name__))
+
+
+    def do_paginate(self, cursor, index):
+        page = request.args.get('page', type=int)
+        if page is None:
+            page = 1
+        elif page == 1 and not request.is_xhr:
+            return None, self.redirect(index)
+
+        action = request.form.get('action') or request.args.get('action')
+        if request.is_xhr and request.path.endswith('.json'):
+            perpage = request.form.get('perpage', type=int) or request.args.get('perpage', type=int) or 999
+        else:
+            perpage = request.form.get('perpage', type=int) or request.args.get('perpage', type=int) or self.perpage
+        limit = request.form.get('limit', False, type=bool) or request.args.get('perpage', type=bool)
+
+        if action == 'paginate':
+            if perpage:
+                self.session['perpage'] = perpage
+            elif perpage == 0:
+                self.session.pop('perpage')
+        else:
+            # Use session
+            perpage = self.session.get('perpage', perpage)
+
+        paginator = Paginator(cursor, page, perpage)
+        if limit:
+            paginator.show = False
+
+        if paginator.pages and page > paginator.pages:
+            return None, self.redirect(index, page=paginator.pages)
+        return paginator, None
+
+
+    def do_index(self, paginator, index):
+        raise Exception('Define `{}.do_index` method'.format(self.__class__.__name__))
+
+
+    def create_model(self):
+        raise Exception('Define `{}.create_model` method'.format(self.__class__.__name__))
+
+
+    def edit_model(self, id):
+        raise Exception('Define `{}.edit_model` method'.format(self.__class__.__name__))
+
+
+    def detail_model(self, id):
+        raise Exception('Define `{}.detail_model` method'.format(self.__class__.__name__))
+
+
+    def do_create(self, model, form):
+        raise Exception('Define `{}.do_create` method'.format(self.__class__.__name__))
+
+
+    def do_edit(self, model):
+        raise Exception('Define `{}.do_edit` method'.format(self.__class__.__name__))
+
+
+    def do_delete(self, models):
+        raise Exception('Define `{}.do_delete` method'.format(self.__class__.__name__))
+
+
+    def do_archive(self, models):
+        raise Exception('Define `{}.do_archive` method'.format(self.__class__.__name__))
+
+
     # HELPERS ------------------------------------------------------------------
     def allows(self, action, models=[], with_filters=True):
         return g.user.allowed(
@@ -299,19 +468,3 @@ class ModelView(View):
         return Response(
             json.dumps(dict(*args, **kwargs)), mimetype='application/json'
         )
-
-
-
-def dashify(text):
-    result = text[0].lower()
-    for c in text[1:]:
-        if c in string.ascii_lowercase:
-            result = result + c
-        else:
-            result = result + '-' + c.lower()
-    return result
-
-
-def utcnow():
-    now = datetime.datetime.utcnow()
-    return now.replace(tzinfo=pytz.utc)
